#!/usr/bin/python

#
# PyCaster - a simple ray casting engine
# by Harry Cheung (http://www.magicookie.com/hacks/pycaster/)
# 

import sys, math
import pygame, pygame.draw, pygame.image
from pygame.locals import *

WIDTH = 640
HEIGHT = 480
ANGLEUNIT = 1024
FOV = ANGLEUNIT / 6
UNIT = 65536
UNITLESS = (UNIT * 8) - 1
RAYANGLE = float(FOV) / WIDTH
HALFHEIGHT = HEIGHT / 2

KEY_UP = K_e
KEY_DOWN = K_d
KEY_LEFT = K_s
KEY_RIGHT = K_f
KEY_RLEFT = K_j
KEY_RRIGHT = K_k

SINTABLE = []
COSTABLE = []
TANTABLE = []
ISINTABLE = []
ICOSTABLE = []
ITANTABLE = []
XUNITTABLE = []
YUNITTABLE = []

for degree in range(ANGLEUNIT):
  radian = degree * math.pi / (ANGLEUNIT / 2)
  SINTABLE.append(math.sin(radian))
  COSTABLE.append(math.cos(radian))
  try:
    ISINTABLE.append(1.0 / math.sin(radian))
  except ZeroDivisionError:
    ISINTABLE.append(0)
  try:
    ICOSTABLE.append(1.0 / math.cos(radian))
  except ZeroDivisionError:
    ICOSTABLE.append(0)

for degree in range((ANGLEUNIT / 4) + 1):
  radian = degree * math.pi / (ANGLEUNIT / 2)
  TANTABLE.append(math.tan(radian))
  try:
    ITANTABLE.append(1.0 / math.tan(radian))
    XUNITTABLE.append(int(round(UNIT * ITANTABLE[-1])))
    YUNITTABLE.append(int(round(UNIT * TANTABLE[-1])))
  except Exception:
    ITANTABLE.append(0)
    XUNITTABLE.append(0)
    YUNITTABLE.append(0)

VIEWDISTANCE = (WIDTH / 2) / TANTABLE[int(FOV / 2)]
UNITVIEWDISTANCE = UNIT * VIEWDISTANCE

class Point:
  def __init__(self, x = 0, y = 0):
    self.x = x
    self.y = y

def horizon_angle(angle):
  if angle <= (ANGLEUNIT / 4):
    return angle
  elif angle <= (ANGLEUNIT / 2):
    return (ANGLEUNIT / 2) - angle
  elif angle <= (ANGLEUNIT * 3 / 4):
    return angle - (ANGLEUNIT / 2)
  else:
    return ANGLEUNIT - angle

angle45 = (ANGLEUNIT / 8)

def distance(ax, ay, bx, by, angle):
  if angle > angle45:
    return math.fabs(ay - by) * ISINTABLE[angle]
  else:
    return math.fabs(ax - bx) * ICOSTABLE[angle]

class View:
  def __init__(self):
    self.screen = pygame.display.set_mode((WIDTH, HEIGHT),
                                          HWSURFACE | DOUBLEBUF)
    self.back_buffer = pygame.Surface(self.screen.get_size()).convert()
    pygame.display.set_caption('PyCaster')
    self.angles = []
    for column in range(WIDTH):
      self.angles.append((column, int(FOV / 2 - round(RAYANGLE * column))))
    self.quadrant = []
    for angle in range(ANGLEUNIT):
      self.quadrant.append((angle / (ANGLEUNIT / 4) + 1, horizon_angle(angle)))

  def set_level(self, level):
    self.position = level.get_start_position()
    self.angle = level.get_start_angle()
    self.grid = level.get_grid()

  def draw(self):
    # draw the ceiling
    pygame.draw.rect(self.back_buffer, (100, 100, 150),
                     (0, 0, WIDTH, HALFHEIGHT), 0)
    # draw the floor
    pygame.draw.rect(self.back_buffer, (0, 0, 0),
                     (0, HALFHEIGHT, WIDTH, HALFHEIGHT), 0)

    # raycast
    hx = 0
    hy = 0
    vx = 0
    vy = 0
    unitposx = (self.position.x / UNIT) * UNIT
    unitposy = (self.position.y / UNIT) * UNIT
    collfn = self.collision
    selfangle = self.angle
    selfquadrant = self.quadrant
    selfposy = self.position.y
    selfposx = self.position.x
    for (column, angle) in self.angles:
      angle = (selfangle + angle) % ANGLEUNIT
      #if column == 0: print 'column:', column
      distort_angle = angle - selfangle
      if distort_angle < 0:
        distort_angle = -distort_angle
      (quadrant, quad_angle) = selfquadrant[angle]
      #if column == 0: print 'quadrant:', quadrant
      #if column == 0: print 'quad_angle:', quad_angle
      #if column == 0: print 'angle:', angle
      ydirect = 1
      hxdelta = UNIT
      hydelta = UNIT
      xdirect = 1
      vxdelta = UNIT
      vydelta = UNIT    
      if quad_angle == 0:
        vx = -1
        hydelta = 0
        hxdelta = UNIT
        if quadrant == 2 or quadrant == 3:
          hx = unitposx - 1
          xdirect = -1
          hxdelta = -hxdelta
        else:
          hx = unitposx + UNIT
        hy = selfposy
      elif quad_angle == (ANGLEUNIT / 4):
        vx = -1
        hydelta = UNIT
        hxdelta = 0
        if quadrant < 3:
          hy = unitposy - 1
          ydirect = -1
          hydelta = -hydelta
        else:
          hy = unitposy + UNIT
        hx = selfposx
      else:
        tanangle = TANTABLE[quad_angle]
        itanangle = ITANTABLE[quad_angle]
        if quadrant == 2 or quadrant == 3:
          xdirect = -1
          vxdelta = -vxdelta
        if quadrant < 3:
          hy = unitposy - 1
          hx = selfposx + \
                   int(round(xdirect * (selfposy - hy) * itanangle))
          ydirect = -1
          hydelta = -hydelta
        else:
          hy = unitposy + UNIT
          hx = selfposx + \
                   int(round(xdirect * (hy - selfposy) * itanangle))
        hxdelta = xdirect * XUNITTABLE[quad_angle]

        if quadrant == 2 or quadrant == 3:
          vx = unitposx - 1
          vy = selfposy + \
                   int(round(ydirect * (selfposx - vx) * tanangle))
        else:
          vx = unitposx + UNIT
          vy = selfposy + \
                   int(round(ydirect * (vx - selfposx) * tanangle))
        vydelta = ydirect * YUNITTABLE[quad_angle]
        #if column == 0: print 'h:', (hx, hy)
        #if column == 0: print 'v:', (vx, vy)
        #if column == 0: print 'hstep:', (hxdelta, hydelta)
        #if column == 0: print 'vstep:', (vxdelta, vydelta)
      hcoll = 0
      while not (hx < 0 or hx > UNITLESS or hy < 0 or hy > UNITLESS):
        #if column == 0: print 'h:', (hx, hy)
        hcoll = collfn(hx, hy, quadrant)
        if hcoll:
          break
        hx = hx + hxdelta
        hy = hy + hydelta
      vcoll = 0
      while not (vx < 0 or vx > UNITLESS or vy < 0 or vy > UNITLESS):
        #if column == 0: print 'v:', (vx, vy)
        vcoll = collfn(vx, vy, quadrant)
        if vcoll:
          break
        vx = vx + vxdelta
        vy = vy + vydelta
      testpoint = None
      hvalid = not (hx < 0 or hx > UNITLESS or hy < 0 or hy > UNITLESS)
      vvalid = not (vx < 0 or vx > UNITLESS or vy < 0 or vy > UNITLESS)
      #if column == 0: print 'hv:', hvalid, 'vv:', vvalid
      #if column == 0: print 'hc:', hcoll, 'vc:', vcoll
      dist = 0
      hdist = 0
      vdist = 0
      coll = 0
      if hvalid:
        hdist = distance(selfposx, selfposy, hx, hy, quad_angle)
      if vvalid:
        vdist = distance(selfposx, selfposy, vx, vy, quad_angle)
      #if column == 0: print 'hd:', hdist, 'vd:', vdist
      if hvalid:
        if vvalid:
          if hdist < vdist:
            dist = hdist
            coll = hcoll
          else:
            dist = vdist
            coll = vcoll
        else:
          dist = hdist
          coll = hcoll
      else:
        dist = vdist
        coll = vcoll
      if distort_angle:
        dist = dist * COSTABLE[distort_angle]
      #if column == 0: print 'dist_angle:', distort_angle
      #if column == 0: print 'dist:', dist
      try:
        height = UNITVIEWDISTANCE / dist
      except ZeroDivisionError:
        print 'angle:', angle
        print 'column:', column
        print 'dist:', dist
        raise ZeroDivisionError
      #if column == 0: print 'height:', height
      #if column == 0:
        pygame.draw.line(self.back_buffer, (100, 100, 100),
                         (column, 0), (column, HEIGHT))
      else:
        if coll == 1:
          self.draw_line(self.back_buffer, (100, 255, 100),
                    (column, HALFHEIGHT - height / 2),
                    (column, HALFHEIGHT + height / 2))
        else:
          self.draw_line(self.back_buffer, (255, 100, 100),
                    (column, HALFHEIGHT - height / 2),
                    (column, HALFHEIGHT + height / 2))
    self.screen.blit(self.back_buffer, (0, 0))
    pygame.display.flip()

  def draw_line(self, a, b, c, d):
    pygame.draw.line(a, b, c, d)

  def collision(self, px, py, quadrant):
    return self.grid[px / UNIT][py / UNIT]

  def move_y(self, unit):
    self.position.x = self.position.x + unit * int(.25 * UNIT * COSTABLE[self.angle])
    self.position.y = self.position.y - unit * int(.25 * UNIT * SINTABLE[self.angle])

  def move_x(self, unit):
    angle = (self.angle - (ANGLEUNIT / 4)) % ANGLEUNIT
    self.position.x = self.position.x + unit * int(.25 * UNIT * COSTABLE[angle])
    self.position.y = self.position.y - unit * int(.25 * UNIT * SINTABLE[angle])

  def rotate(self, unit):
    self.angle = (self.angle + unit) % ANGLEUNIT

class Level:
  def __init__(self):
    self.start_position = Point(262144, 352846)
    self.start_angle = (ANGLEUNIT / 2)
    self.grid = [[1, 2, 1, 2, 1, 2, 1, 2],
                 [2, 0, 0, 0, 0, 0, 0, 1],
                 [1, 0, 0, 0, 0, 0, 0, 2],
                 [2, 0, 0, 1, 2, 0, 0, 1],
                 [1, 0, 0, 2, 1, 0, 0, 2],
                 [2, 0, 0, 0, 0, 0, 0, 1],
                 [1, 0, 0, 0, 0, 0, 0, 2],
                 [2, 1, 2, 1, 2, 1, 2, 1]]                 

  def get_start_position(self):
    return self.start_position

  def get_start_angle(self):
    return self.start_angle

  def get_grid(self):
    return self.grid

def main(argv):
  pygame.init()
  pygame.event.set_blocked([ACTIVEEVENT, MOUSEBUTTONUP,
                            MOUSEBUTTONDOWN, JOYAXISMOTION, JOYBALLMOTION,
                            JOYHATMOTION, JOYBUTTONUP, JOYBUTTONDOWN,
                            VIDEORESIZE, VIDEOEXPOSE])
  view = View()
  view.set_level(Level())

  down_key = 0
  last_time = pygame.time.get_ticks() - 1.0
  start_time = pygame.time.get_ticks()
  moving_y = 0
  moving_x = 0
  rotating = 0
  frames = 0
  while 1:
    event = pygame.event.poll()

    if event.type == QUIT:
      break
    elif event.type == KEYDOWN:
      if event.key == K_ESCAPE:
        break
      elif event.key == K_F1:
        pygame.image.save(view.screen, 'screenshot.bmp')
      elif event.key == K_r:
        print 'view angle:', view.angle
        print 'view pos:', view.position
      elif event.key == KEY_UP:
        moving_y = 1
      elif event.key == KEY_DOWN:
        moving_y = -1
      elif event.key == KEY_LEFT:
        moving_x = -1
      elif event.key == KEY_RIGHT:
        moving_x = 1
      elif event.key == KEY_RRIGHT:
        rotating = -(ANGLEUNIT / 72)
      elif event.key == KEY_RLEFT:
        rotating = (ANGLEUNIT / 72)
    elif event.type == KEYUP:
      if event.key == KEY_UP and moving_y == 1:
        moving_y = 0
      elif event.key == KEY_DOWN and moving_y == -1:
        moving_y = 0
      elif event.key == KEY_LEFT and moving_x == -1:
        moving_x = 0
      elif event.key == KEY_RIGHT and moving_x == 1:
        moving_x = 0
      elif event.key == KEY_RLEFT and rotating == (ANGLEUNIT / 72):
        rotating = 0
      elif event.key == KEY_RRIGHT and rotating == -(ANGLEUNIT / 72):
        rotating = 0
    elif event.type == MOUSEMOTION:
      if event.pos[0] > (WIDTH / 2 + 50):
        rotating = -(ANGLEUNIT / 72)
      elif event.pos[0] < (WIDTH / 2 - 50):
        rotating = (ANGLEUNIT / 72)
      else:
        rotating = 0
    
    if moving_y != 0:
      view.move_y(moving_y)
    if moving_x != 0:
      view.move_x(moving_x)
    if rotating != 0:
      view.rotate(rotating)

    current_time = pygame.time.get_ticks()
    last_time = current_time
    frames = frames + 1
    view.draw()
  print 'fps:', (frames * 1000.0) / (pygame.time.get_ticks() - start_time)

if __name__ == '__main__':
  import sys
  main(sys.argv)
